""" Widgets for mongonaut forms"""

from django import forms

from mongoengine.base import ObjectIdField
from mongoengine.fields import BooleanField
from mongoengine.fields import DateTimeField
from mongoengine.fields import EmbeddedDocumentField
from mongoengine.fields import ListField
from mongoengine.fields import ReferenceField


def get_widget(field, disabled=False):
    
    if isinstance(field, ListField) or \
        isinstance(field, EmbeddedDocumentField):
        return None
    
    attrs={}
    attrs['class'] = 'span6 xlarge'
    if disabled or isinstance(field, ObjectIdField):
        attrs['class'] += ' disabled'
        attrs['readonly'] = 'readonly'
    
    if hasattr(field, "max_length") and not field.max_length:
        return forms.Textarea(attrs=attrs)

    if isinstance(field, DateTimeField):
        return forms.DateTimeInput(attrs=attrs)
        
    if isinstance(field, BooleanField):
        return forms.CheckboxInput(attrs=attrs)
        
    if isinstance(field, ReferenceField):
        return forms.Select(attrs=attrs)

    return forms.TextInput(attrs=attrs)


from django.conf.urls.defaults import patterns, include, url
from django.views.generic import ListView

from articles.models import Post

urlpatterns = patterns('',
    url(
        regex=r'^$',
        view=ListView.as_view(
            queryset=Post.objects.all(),
            template_name="articles/post_list.html"
        ),
        name="article_list"
    ),
)



from django.views.generic import ListView

from articles.models import Post


class PostListView(ListView):
    
    template_name="articles/post_list.html"
    
    def get_queryset(self):
        
        return Post.objects.all()



__author__ = 'Daniel Greenfeld'

VERSION = (0, 2, 11)

def get_version():
    version = '%s.%s' % (VERSION[0], VERSION[1])
    if VERSION[2]:
        version = '%s.%s' % (version, VERSION[2])
    return version

__version__ = get_version()


"""
    TODO move permission checks to the dispatch view thingee
"""

from datetime import datetime

from django.conf import settings
from django.contrib import messages
from django.core.urlresolvers import reverse
from django.forms import widgets
from django.forms.widgets import DateTimeInput, CheckboxInput
from django.http import HttpResponseForbidden, HttpResponseRedirect
from django.views.generic.edit import DeletionMixin
from django.views.generic import DetailView
from django.views.generic import ListView
from django.views.generic import TemplateView
from django.views.generic.edit import FormView

from mongoengine.fields import EmbeddedDocumentField, ListField

from mongonaut.forms import DocumentListForm
from mongonaut.forms import DocumentDetailForm
from mongonaut.forms import document_detail_form_factory
from mongonaut.mixins import MongonautViewMixin
from mongonaut.utils import is_valid_object_id

class IndexView(MongonautViewMixin, ListView):

    template_name = "mongonaut/index.html"
    queryset = []
    
    def get_queryset(self):
        return self.get_mongoadmins()

class AppListView(MongonautViewMixin, ListView):
    """ :args: <app_label> """

    template_name = "mongonaut/app_list.html"

class DocumentListView(MongonautViewMixin, FormView):
    """ :args: <app_label> <document_name> 
    
        TODO - Make a generic document fetcher method
    """
    form_class = DocumentListForm
    success_url = '/'
    template_name = "mongonaut/document_list.html"
    
    #def dispatch(self, *args, **kwargs):
    #    self.set_mongoadmin()
    #    self.set_permissions()
    #    return super(DocumentListView, self).dispatch(*args, **kwargs)    
    
    def get_qset(self, queryset, q):
        if self.mongoadmin.search_fields and q:
            params = {}
            for field in self.mongoadmin.search_fields:
                if field == 'id':
                    # check to make sure this is a valid ID, otherwise we just continue
                    if is_valid_object_id(q):
                        return queryset.filter(id=q)
                    continue
                search_key = "{field}__icontains".format(field=field)
                params[search_key] = q

            queryset = queryset.filter(**params)        
        return queryset
    
    def get_queryset(self):
        if hasattr(self, "queryset") and self.queryset:
            return self.queryset
        
        self.set_mongonaut_base()
        self.set_mongoadmin()        
        self.document = getattr(self.models, self.document_name)
        queryset = self.document.objects.all()
        
        # search. move this to get_queryset        
        # search. move this to get_queryset
        q = self.request.GET.get('q')
        query = self.get_qset(queryset, q)

        
        ### Start pagination
        ### Note: 
        ###    Didn't use the Paginator in Django cause mongoengine querysets are 
        ###    not the same as Django ORM querysets and it broke.
        # Make sure page request is an int. If not, deliver first page.
        try:
            page = int(self.request.GET.get('page', '1'))
        except ValueError:
            page = 1
        self.page = page
            
        self.documents_per_page = 25
        self.total_pages = max(queryset.count() / self.documents_per_page, 1)
        start = (self.page -1) * self.documents_per_page
        end = self.page * self.documents_per_page
        self.previous_page_number = page - 1        
        self.next_page_number = page + 1
        try:
            queryset = queryset[start:end]
        except Exception as e:
            print e

        self.queryset = queryset
        return queryset
        
    def get_initial(self):
        self.query = self.get_queryset()
        mongo_ids = {'mongo_id':[unicode(x.id) for x in self.query]}
        return mongo_ids
        
    def get_context_data(self, **kwargs):
        context = super(DocumentListView, self).get_context_data(**kwargs)
        context = self.set_permissions_in_context(context)
        if not context['has_view_permission']:
            return HttpResponseForbidden("You do not have permissions to view this content.")
        
        context['object_list'] = self.get_queryset()        
        context['document'] = self.document
        context['app_label'] = self.app_label  
        context['document_name'] = self.document_name
        
        # pagination bits
        context['page'] = self.page
        context['documents_per_page'] = self.documents_per_page
        context['previous_page_number'] = self.previous_page_number
        context['next_page_number'] = self.next_page_number
        context['total_pages'] = self.total_pages

        # Part of upcoming list view form functionality
        if self.queryset.count():
            context['keys'] = ['id',]
            
            # Show those items for which we've got list_fields on the mongoadmin
            for key in [x for x in self.document._fields.keys() if x != 'id' and x in self.mongoadmin.list_fields]:
                
                # TODO - Figure out why this EmbeddedDocumentField and ListField breaks this view
                # Note - This is the challenge part, right? :)
                if isinstance(self.document._fields[key], EmbeddedDocumentField):            
                    continue
                if isinstance(self.document._fields[key], ListField):                                
                    continue
                context['keys'].append(key)
        
        if self.mongoadmin.search_fields:
            context['search_field'] = True

        return context

    def post(self, request, *args, **kwargs):
        # TODO - make sure to check the rights of the poster
        #self.get_queryset() # TODO - write something that grabs the document class better
        form_class = self.get_form_class()
        form = self.get_form(form_class)
        mongo_ids = self.get_initial()['mongo_id']             
        for form_mongo_id in form.data.getlist('mongo_id'):
            for mongo_id in mongo_ids:
                if form_mongo_id == mongo_id:
                    self.document.objects.get(id=mongo_id).delete()
            
        return self.form_invalid(form)                                    

class DocumentDetailView(MongonautViewMixin, TemplateView):
    """ :args: <app_label> <document_name> <id> """
    template_name = "mongonaut/document_detail.html"
    
    def get_context_data(self, **kwargs):
        context = super(DocumentDetailView, self).get_context_data(**kwargs)
        self.set_mongoadmin()        
        context = self.set_permissions_in_context(context)
        if not context['has_view_permission']:
            return HttpResponseForbidden("You do not have permissions to view this content.")
        self.document_type = getattr(self.models, self.document_name)
        self.ident = self.kwargs.get('id')
        self.document = self.document_type.objects.get(id=self.ident)
        
        context['document'] = self.document
        context['app_label'] = self.app_label  
        context['document_name'] = self.document_name
        context['keys'] = ['id',]
        context['embedded_documents'] = []        
        context['list_fields'] = []
        for key in sorted([x for x in self.document._fields.keys() if x != 'id']):
            # TODO - Figure out why this EmbeddedDocumentField and ListField breaks this view
            # Note - This is the challenge part, right? :)
            if isinstance(self.document._fields[key], EmbeddedDocumentField):            
                context['embedded_documents'].append(key)
                continue
            if isinstance(self.document._fields[key], ListField):                                
                context['list_fields'].append(key)
                continue                
            context['keys'].append(key)
        return context
 

class DocumentEditFormView(MongonautViewMixin, FormView):
    """ :args: <app_label> <document_name> <id> """#

    template_name = "mongonaut/document_edit_form.html"
    form_class = DocumentDetailForm
    success_url = '/'
    
    def get_success_url(self):
        self.set_mongonaut_base() 
        return reverse('document_detail_edit_form', kwargs={'app_label':self.app_label,'document_name':self.document_name,'id':self.kwargs.get('id')})
    
    def get_context_data(self, **kwargs):
        context = super(DocumentEditFormView, self).get_context_data(**kwargs)
        self.set_mongoadmin()        
        context = self.set_permissions_in_context(context)
        if not context['has_edit_permission']:
            return HttpResponseForbidden("You do not have permissions to edit this content.")
        self.document_type = getattr(self.models, self.document_name)
        self.ident = self.kwargs.get('id')
        self.document = self.document_type.objects.get(id=self.ident)        
        
        context['document'] = self.document
        context['app_label'] = self.app_label
        context['document_name'] = self.document_name

        return context
        
    def get_form(self, DocumentDetailForm):
        self.set_mongoadmin()
        context = self.set_permissions_in_context({})
        if not context['has_edit_permission']:
            return HttpResponseForbidden("You do not have permissions to edit this content.")
        
        self.document_type = getattr(self.models, self.document_name)
        self.ident = self.kwargs.get('id')
        self.document = self.document_type.objects.get(id=self.ident)
        self.form = DocumentDetailForm()                
            
        self.form = document_detail_form_factory(form=self.form, document_type=self.document_type, initial=self.document)
        if self.request.method == 'POST':
            self.form.data = self.request.POST
            self.form.is_bound = True
            if self.form.is_valid():
                for key, field in self.form.fields.items():                      
                    if 'readonly' in field.widget.attrs:
                        # For _id or things specified as such
                        continue       
                        
                    if isinstance(field.widget, DateTimeInput):
                        format = field.widget.format
                        setattr(self.document, key, datetime.strptime(self.request.POST[key], format))
                        continue


                    if isinstance(field.widget, CheckboxInput):
                        if key in self.request.POST:
                            setattr(self.document, key, True)
                        else:
                            setattr(self.document, key, False)
                        continue
                        
                    if isinstance(field.widget, widgets.Select):
                        # supporting reference fields!
                        value = field.mongofield.document_type.objects.get(id=self.request.POST[key])
                        setattr(self.document, key, value)
                        continue

                    # for strings
                    setattr(self.document, key, self.request.POST[key])
                    
                self.document.save()
                messages.add_message(self.request, messages.INFO, 'Your changes have been saved.')
                
        return self.form
        
class DocumentAddFormView(MongonautViewMixin, FormView):
    """ :args: <app_label> <document_name> <id> """#

    template_name = "mongonaut/document_add_form.html"
    form_class = DocumentDetailForm
    success_url = '/'

    def get_success_url(self):
        self.set_mongonaut_base()  
        return reverse('document_detail', kwargs={'app_label':self.app_label,'document_name':self.document_name,'id':str(self.document.id)})

    def get_context_data(self, **kwargs):
        """ TODO - possibly inherit this from DocumentEditFormView. This is same thing minus:
            self.ident = self.kwargs.get('id')
            self.document = self.document_type.objects.get(id=self.ident)
        """
        context = super(DocumentAddFormView, self).get_context_data(**kwargs)
        self.set_mongoadmin()        
        context = self.set_permissions_in_context(context)
        if not context['has_add_permission']:
            return HttpResponseForbidden("You do not have permissions to edit view content.")
        self.document_type = getattr(self.models, self.document_name)
        
        context['app_label'] = self.app_label
        context['document_name'] = self.document_name
        return context

    def get_form(self, DocumentDetailForm):
        self.set_mongonaut_base()
        self.document_type = getattr(self.models, self.document_name)
        self.form = DocumentDetailForm()        
        self.form = document_detail_form_factory(form=self.form, document_type=self.document_type)
        if self.request.method == 'POST':
            self.form.data = self.request.POST
            self.form.is_bound = True
            if self.form.is_valid():
                self.document = self.document_type()              
                for key, field in self.form.fields.items():                      
                    if 'readonly' in field.widget.attrs:
                        # For _id
                        continue       

                    if isinstance(field.widget, DateTimeInput):
                        format = field.widget.format
                        setattr(self.document, key, datetime.strptime(self.request.POST[key], format))
                        continue

                    if isinstance(field.widget, CheckboxInput):
                        if key in self.request.POST:
                            setattr(self.document, key, True)
                        else:
                            setattr(self.document, key, False)
                        continue
                        
                    if isinstance(field.widget, widgets.Select):
                        # supporting reference fields!
                        value = field.mongofield.document_type.objects.get(id=self.request.POST[key])
                        setattr(self.document, key, value)
                        continue                        

                    # for strings
                    setattr(self.document, key, self.request.POST[key])
                self.document.save()
                messages.add_message(self.request, messages.INFO, 'Your new document has been added and saved.')

        return self.form

class DocumentDeleteView(DeletionMixin, MongonautViewMixin, TemplateView):
    """ :args: <app_label> <document_name> <id> 
    
        TODO - implement a GET view for confirmation
    """
    
    success_url = "/"
    template_name = "mongonaut/document_delete.html"
    
    def get_success_url(self):
        self.set_mongonaut_base()
        messages.add_message(self.request, messages.INFO, 'Your document has been deleted.')
        return reverse('document_list', kwargs={'app_label':self.app_label,'document_name':self.document_name})

    def get_object(self):
        self.set_mongoadmin()                
        self.document_type = getattr(self.models, self.document_name)
        self.ident = self.kwargs.get('id')
        self.document = self.document_type.objects.get(id=self.ident)
        return self.document   

class ListFieldListView(MongonautViewMixin, FormView):
    pass

class EmbeddedDocumentView(MongonautViewMixin, DetailView):
    pass


class EmbeddedDocumentDetailView(MongonautViewMixin, DetailView):
    """ :args: <app_label> <document_name> <id> <???> """
    template_name = "mongonaut/embedded_document_detail.html"


from mongoengine.base import ObjectIdField, ValidationError

# Used to valid object_ids.
# Called by is_valid_object_id
OBJECT_ID = ObjectIdField()

def is_valid_object_id(value):
    try:
        OBJECT_ID.validate(value)
        return True
    except ValidationError:
        return False



from mongoengine import BooleanField
from mongoengine import DateTimeField
from mongoengine import Document
from mongoengine import EmbeddedDocument
from mongoengine import EmbeddedDocumentField
from mongoengine import ListField
from mongoengine import ReferenceField
from mongoengine import StringField

class User(Document):
    email = StringField(required=True, max_length=50)
    first_name = StringField(max_length=50)
    last_name = StringField(max_length=50)
    
    def __unicode__(self):
        return self.email

class EmbeddedUser(EmbeddedDocument):
    email = StringField(required=True, max_length=50)
    first_name = StringField(max_length=50)
    last_name = StringField(max_length=50)

    
class Comment(EmbeddedDocument):
    content = StringField()
    name = StringField(max_length=120)    

class Post(Document):
    # See Post.title.max_length to make validation better!
    title = StringField(max_length=120, required=True)
    author = ReferenceField(User)
    content = StringField()    
    tags = ListField(StringField(max_length=30))
    comments = ListField(EmbeddedDocumentField(Comment))    
    pub_date = DateTimeField()
    published = BooleanField()
    creator = EmbeddedDocumentField(EmbeddedUser)    # for testing purposes




# Make our own testrunner that by default only tests our own apps

from django.conf import settings
from django.test.simple import DjangoTestSuiteRunner
from django_coverage.coverage_runner import CoverageRunner

class OurTestRunner(DjangoTestSuiteRunner):
    def build_suite(self, test_labels, *args, **kwargs):
        return super(OurTestRunner, self).build_suite(test_labels or settings.PROJECT_APPS, *args, **kwargs)

class OurCoverageRunner(OurTestRunner, CoverageRunner):
    pass



import os
from setuptools import setup, find_packages
 
import mongonaut
 
LONG_DESCRIPTION = open('README.rst').read()
 
setup(
    name='django-mongonaut',
    version=mongonaut.__version__,
    description="An introspective interface for Django and MongoDB",
    long_description=LONG_DESCRIPTION,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Framework :: Django",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: JavaScript",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords='mongodb,django',
    author=mongonaut.__author__,
    author_email='pydanny@gmail.com',
    url='http://github.com/pydanny/django-mongonaut',
    license='MIT',
    packages=find_packages(),
    include_package_data=True,
    install_requires=['mongoengine==0.5.2'],
    zip_safe=False,
)


from django.conf import settings
from django.utils.importlib import import_module

from django.conf import settings

class AppStore(object):
    
    def __init__(self, module):
        self.models = []
        for key in module.__dict__.keys():
            model_candidate = getattr(module, key)
            if hasattr(model_candidate, 'mongoadmin'):
                self.add_model(model_candidate)
                
    def add_model(self, model):
        model.name = model.__name__
        self.models.append(model)
        
class MongonautViewMixin(object):
    
    def get_context_data(self, **kwargs):
        context = super(MongonautViewMixin, self).get_context_data(**kwargs)
        context['MONGONAUT_JQUERY'] = getattr(settings, "MONGONAUT_JQUERY", "http://ajax.googleapis.com/ajax/libs/jquery/1.7.1/jquery.min.js")
        context['MONGONAUT_TWITTER_BOOTSTRAP'] = getattr(settings, "MONGONAUT_TWITTER_BOOTSTRAP", "http://twitter.github.com/bootstrap/assets/css/bootstrap.css")
        context['MONGONAUT_TWITTER_BOOTSTRAP_ALERT'] = getattr(settings, "MONGONAUT_TWITTER_BOOTSTRAP_ALERT", "http://twitter.github.com/bootstrap/assets/js/bootstrap-alert.js")
        return context
    
    def get_mongoadmins(self):
        """ Returns a list of all mongoadmin implementations for the site """
        apps = []        
        for app_name in settings.INSTALLED_APPS:
            mongoadmin = "{0}.mongoadmin".format(app_name)
            try:
                module = import_module(mongoadmin)
            except ImportError as e:
                if str(e) == "No module named mongoadmin":
                    continue
                raise e
            
            app_store = AppStore(module)
            apps.append(dict(
                app_name=app_name,
                obj=app_store
            ))
        return apps
        
    
    def set_mongonaut_base(self):
        """ Sets a number of commonly used attributes """        
        if hasattr(self, "app_label"):
            # prevents us from calling this multiple times
            return None
        self.app_label = self.kwargs.get('app_label')
        self.document_name = self.kwargs.get('document_name')
    
        # TODO Allow this to be assigned via url variable
        self.models_name = self.kwargs.get('models_name', 'models')
    
        # import the models file
        self.model_name = "{0}.{1}".format(self.app_label, self.models_name)
        self.models = import_module(self.model_name)
        
    def set_mongoadmin(self):
        """ Returns the MongoAdmin object for an app_label/document_name style view
        """
        if hasattr(self, "mongoadmin"):
            return None
                    
        if not hasattr(self, "document_name"):
            self.set_mongonaut_base()
        
        for mongoadmin in self.get_mongoadmins():
            for model in mongoadmin['obj'].models:
                if model.name == self.document_name:
                    self.mongoadmin = model.mongoadmin
                    break
        # TODO change this to use 'finally' or 'else' or something
        if not hasattr(self, "mongoadmin"):
            raise NoMongoAdminSpecified("No MongoAdmin for {0}.{1}".format(self.app_label, self.document_name))

    def set_permissions_in_context(self, context={}):
        """ Provides permissions for mongoadmin for use in the context"""
        
        context['has_view_permission'] = self.mongoadmin.has_view_permission(self.request)
        context['has_edit_permission'] = self.mongoadmin.has_edit_permission(self.request)
        context['has_add_permission'] = self.mongoadmin.has_add_permission(self.request)
        context['has_delete_permission'] = self.mongoadmin.has_delete_permission(self.request)
        return context
            